/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifdef WIN32
#include <windows.h>
#include <shlobj.h>
#include <process.h>
#pragma warning(disable : 4996)
#else
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#endif

#include <stdio.h>
#include <time.h>
#include <string.h>
#include "common.h"

char *AgentName();
char *agentname();

void str_c2c(char *buff, char f, char t)
{
	size_t len = strlen(buff);
	size_t i;
	for (i=0;i<len;++i) 
	{
		if (buff[i]==f)
		{
			buff[i]=t;
		}
	}
}

void bs2s(char *buff)
{
	str_c2c(buff, '\\', '/');
}
void s2bs(char *buff)
{
	str_c2c(buff, '/', '\\');
}


char __dir[MAX_PATH];
char *getdir(void)
{
	return __dir;
}
char __exe[MAX_PATH];
char *getexefile(void)
{
	return __exe;
}



void writemsg(FILE *fd, char *msg, char *syserror)
{
#define DATEFORMAT "%02d-%02d-%02d %02d:%02d:%02d: "
	struct tm *ttp;
        time_t t = time(NULL);
#ifdef WIN32
	ttp = localtime(&t);	
#else
	struct tm tt;
        ttp = localtime_r(&t, &tt);
#endif


	fprintf(fd, syserror==NULL?DATEFORMAT"%s\n":DATEFORMAT"%s [sys: %s]\n",
		ttp->tm_year+1900, ttp->tm_mon+1, ttp->tm_mday, ttp->tm_hour, ttp->tm_min, ttp->tm_sec,
 		msg, syserror);
	fflush(fd);
}

FILE *openlogfile(char *fname) 
{
	FILE *fd = fopen(fname, "r+t");
	if (fd!=NULL)
	{
		int pos;
		fseek(fd, 0, SEEK_END);
		pos = ftell(fd);
		if (pos>8192)
		{
			char buff[256];
			size_t ret;
			fseek(fd, -256, SEEK_END);
			ret = fread(buff,1,256,fd);
			fclose(fd);
			fd = fopen(fname, "wt");
			if (ret>0)
			{
				fprintf(fd, "...");
				fwrite(buff, 1,ret,fd);
			}
		}

	}
	else
	{
		fd = fopen(fname, "at");
	}
	return fd;
}


void __logmsg(char *msg, char *syserror)
{
	char fname[512];
	FILE *fd;

	writemsg(stdout, msg, syserror);

	sprintf(fname, "%s%s.log", getdir(), agentname() );
	fd = openlogfile(fname);
	if (fd!=NULL)
	{
		writemsg(fd, msg, syserror);
		fclose(fd);
	}
}

void iao_logmsg(char *msg)
{
	__logmsg(msg, NULL);
}

void syserror(char *buff, int errcode)
{
#ifdef WIN32
	LPVOID lpMsgBuf;
	size_t len;
	if (FormatMessageA( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		errcode,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),/* Default language */
		(LPSTR) &lpMsgBuf,
		0,
		NULL 
		)>0)
	{
		len = strlen(lpMsgBuf);
		len = len>1000?999:len-2;
		strncpy(buff, (char*)lpMsgBuf, len);
		buff[len]=0;
		LocalFree( lpMsgBuf );
	}
	else
	{
		strcpy(buff, "ext. error info failed");
	}
#else
	strcpy(buff, strerror(errcode));
#endif
}
int syserrorcode(void)
{
#ifdef WIN32
	return GetLastError();
#else
	return errno;
#endif
}

void logsysmsgforce(char *msg, int errcode)
{
	char buff[1001];
	syserror(buff, errcode);
	__logmsg(msg, buff);
}
void iao_logsysmsg(char *msg)
{
	logsysmsgforce(msg, syserrorcode());
}



int copyfile(char *src, char *dst)
{
	char buff[4096];
	size_t ret; 
	FILE *d;
	FILE *s = fopen(src, "rb");
	if (s==NULL)
	{
		iao_logsysmsg("copyfile: Can't open src file");
		return -1;
	}
	d = fopen(dst, "wb");
	if (d==NULL)
	{
		iao_logsysmsg("copyfile: Can't open dst file");
		fclose(s);
		return -1;
	}

	while (!feof(s))
	{
		size_t wr;
		ret = fread(buff, 1, sizeof(buff), s);
		wr = fwrite(buff, 1, ret, d);
		if (wr!=ret)
		{
			iao_logsysmsg("copyfile: Write failed");
			fclose(s);
			fclose(d);
			return -1;
		}
	}

	fclose(s);
	fclose(d);
	return 0;
}

int spawn(char *path)
{
#ifdef WIN32
	char buff[MAX_PATH];
	STARTUPINFO startupinfo = {
		sizeof(STARTUPINFO), NULL, NULL, NULL, 0, 0, 0, 0, 0, 0, 0, 0, SW_HIDE, 0, NULL, NULL, NULL, NULL};
	PROCESS_INFORMATION processinfo;

	sprintf(buff, "\"%s\"", path);

	if (CreateProcess(NULL, path, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &startupinfo, &processinfo)==TRUE)
	{
		CloseHandle(processinfo.hProcess);
		CloseHandle(processinfo.hThread);
		return 0;
	}
	else
	{
		iao_logsysmsg("span: Spawn failed");
		return -1;
	}

	/*if (_spawnlp(_P_NOWAIT, path, buff, NULL)==-1)*/
#else
	char buff[MAX_PATH];
	sprintf(buff, "%s &", path);
	return system(buff);
#endif
}

int execute(char *path)
{
#ifdef WIN32
	char buff[MAX_PATH];
	sprintf(buff, "\"%s\"", path);
	_execlp(path, buff, NULL);
#else
	/* The fork is necessary for Darwin */
	switch (fork())
	{
		case 0:
			execlp(path, path, NULL);
			break;
		case -1:
			break;
		default:
			exit(0);
	}
			
#endif
	iao_logsysmsg("execute: Failed");
	return -1;
}

int ask(char *question)
{
#ifdef WIN32
	return MessageBoxA(NULL, question, AgentName(), MB_YESNO)==IDYES;
#else
	return 0;
#endif
}

void showlog(void)
{
#ifdef WIN32
	char fname[MAX_PATH];
	sprintf(fname, "%s%s.log", getdir(), agentname());
	_execlp("notepad.exe", "notepad.exe", fname, NULL);
#endif
}

void bailout(void)
{
#ifdef WIN32
	showlog();
#endif
	exit(-1);
}

#ifdef WIN32
void syncmsgbox(char *text)
{
        MessageBoxA(NULL, text, AgentName() , MB_ICONINFORMATION|MB_OK|MB_SYSTEMMODAL);
}

DWORD WINAPI t_syncmsgbox(LPVOID text)
{
		syncmsgbox(text);
        return 0;
}


void asyncmsgbox(char *text)
{
        DWORD tid; 
        HANDLE h = CreateThread(NULL, 16384, t_syncmsgbox, (void*)text, 0, &tid);
        if (h==NULL)
        {
                syncmsgbox(text);
        }                                                                  
        else
        {
                CloseHandle(h);
        }
}
#endif


void informok(void)
{
	iao_logmsg("Agent installed successfuly");
#ifdef WIN32
	asyncmsgbox("Agent installed successfuly");
#endif
}

