#include <stdlib.h>
#include <stdio.h>
#include "iao_threads.h"


void *funcandarg(void *f, void *arg)
{
	void **farg = (void**)malloc(sizeof(void*)*2);
	farg[0] = f;
	farg[1] = arg;
	return farg;
}

void callfarg(void **farg)
{
	void (*threadroutine)(void*) = (void(*)(void*)) farg[0];
	void *arg = farg[1];
	free(farg);
	threadroutine(arg);
}

#ifndef WIN32
#include <pthread.h>
#include <stdlib.h>

static  void *threadlauncher(void *arg)
{
	callfarg(arg);
	return NULL;
}

iao_thread iao_spawnthread(void (*threadroutine)(void*), void *arg)
{
	pthread_t t;
	void **farg = funcandarg((void*)threadroutine, arg);

	if (pthread_create(&t, NULL, threadlauncher, farg)==0)
	{
		return (iao_thread) t;
	}
	else
	{
		iao_logsysmsg("iao_spawnthread: Failed creating thread");
		return NULL;
	}
}
iao_thread iao_spawnthread_nojoin(void (*threadroutine)(void*), void *arg)
{
	return iao_spawnthread(threadroutine, arg);
}

int iao_jointhread(iao_thread thread)
{
	void *ret;
	if (pthread_join((pthread_t)thread, &ret)==0)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("iao_jointhread: Failed joining thread");
		return -1;
	}
}


iao_mutex iao_createmutex()
{
	pthread_mutex_t *mutex =
		 (pthread_mutex_t*) malloc(sizeof(pthread_mutex_t));

	if (pthread_mutex_init(mutex, NULL)==0)
	{
		return (iao_mutex) mutex;
	}
	else
	{
		iao_logsysmsg("iao_createmutex: Failed creating mutex");
		return NULL;
	}
}

int iao_lockmutex(iao_mutex mutex)
{
	if (pthread_mutex_lock( (pthread_mutex_t *) mutex)==0)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("iao_lockmutex: Failed locking mutex");
		return -1;
	}
}

int iao_releasemutex(iao_mutex mutex)
{
	if (pthread_mutex_unlock( (pthread_mutex_t*) mutex)==0)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("iao_releasemutex: Failed releasing mutex");
		return -1;
	}
}

int iao_deletemutex(iao_mutex mutex)
{
	if (pthread_mutex_destroy( (pthread_mutex_t*) mutex)==0)
	{
		free(mutex);
		return 0;
	}
	else
	{
		iao_logsysmsg("iao_deletemutex: Failed destroying mutex");
		return -1;
	}
}


#else
#include <windows.h>

static  DWORD WINAPI threadlauncher(void *arg)
{
	callfarg(arg);
	return 0;
}

iao_thread iao_spawnthread(void (*threadroutine)(void*), void *arg)
{
	HANDLE t;

	void **farg = funcandarg(threadroutine, arg);

	t = CreateThread(NULL, 0,  threadlauncher, farg, 0, NULL);
	if (t!=NULL)
	{
		return (iao_thread) t;
	}
	else
	{
		iao_logsysmsg("iao_spawnthread: Failed creating thread");
		return NULL;
	}
}

iao_thread iao_spawnthread_nojoin(void (*threadroutine)(void*), void *arg)
{
	iao_thread thread = iao_spawnthread(threadroutine, arg);
	if (thread!=NULL)
	{
		CloseHandle((HANDLE)thread);
	}	
	return thread;
}

int iao_jointhread(iao_thread thread)
{
	void *ret = NULL;
	if (WaitForSingleObject((HANDLE)thread, INFINITE)==WAIT_OBJECT_0)
	{
		CloseHandle((HANDLE)thread);
		return 0;
	}
	else
	{
		iao_logsysmsg("Failed joining thread");
		return -1;
	}
}


iao_mutex iao_createmutex()
{
	HANDLE mutex = CreateMutex(NULL, FALSE, NULL);
	if (mutex!=NULL)
	{
		return (iao_mutex) mutex;
	}
	else
	{
		iao_logsysmsg("Failed creating mutex");
		return NULL;
	}
}

int iao_lockmutex(iao_mutex mutex)
{
	if (WaitForSingleObject((HANDLE)mutex, INFINITE)==WAIT_OBJECT_0)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("Failed locking mutex");
		return -1;
	}
}

int iao_releasemutex(iao_mutex mutex)
{
	if (ReleaseMutex((HANDLE) mutex)==TRUE)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("Failed releasing mutex");
		return -1;
	}
}

int iao_deletemutex(iao_mutex mutex)
{
	if (CloseHandle((HANDLE)mutex)==TRUE)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("Failed destroying mutex");
		return -1;
	}
}

#endif /* WIN32 */



#ifdef THREAD_TEST
#include <stdio.h>
char *agentname() { return "TEST"; }
void *mutex;
int counter = 10000000;

void thread(void *arg)
{
	int iter = 0;

	int *var = (int*) arg;
	while (*var>0)
	{
		iao_lockmutex(mutex);
		*var = *var - 1;
		iao_releasemutex(mutex);
		iter++;
	}
	printf("iter: %d\n", iter);
}

void *t[3];

int main()
{
	int i;
	mutex = iao_createmutex();
	for (i=0;i<3;++i)
	{
		t[i] = iao_spawnthread(thread, &counter);
	}
	for (i=0;i<3;++i)
	{
		int ret = iao_jointhread(t[i]);
		printf("%d\n", ret);
	}
	iao_deletemutex(mutex);
	return 0;
}
#endif
