/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#include "ciao.h"
#include "cfw.h"
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#include "system.h"

#ifdef WIN32
#pragma warning(disable:4996)
#endif

#define MAX_VOLUMES 50
#define CONSECUTIVEREADINGS 2
#define MAX_ALARMS 256

static char ignorevolumes[4096];

typedef struct metricstatestruct {
	char *name;
	int alarmed;
	int state;
	int threshold;
	int consecutive;
	int value;
} metricstate;

static char *metricnames[] = {
        "CPU Usage", "Physical Memory", "Swap"
};
static char *metricalarmid[] = {
        "CPU_USAGE", "PMEM_USAGE", "SWAP_CAPACITY"
};
static char *metricresourcename[] = {
        "CPU", "Physical Memory", "Swap File"
};
static char *metricdefaultthreshold[] = {
        "90", "95", "85"
};
static char metricseverity[] = {'M', 'M', 'M'};

static metricstate metricstates[METRICSCOUNT];

static time_t lastload;
static volume volumes[MAX_VOLUMES];
static int vols = 0;
int volume_threshold = 85;

void initvolumes()
{
	lastload = time(NULL);
	vols = loadvolumes(volumes, MAX_VOLUMES);
}

int agentinit(iao_agentinfo *agent)
{
	return 60;
}


int reportstate(iao_agentinfo *agent, char *verb, char *x)
{
    iao_alarm alarms[MAX_ALARMS];
	int alarmcount = 0;
	int i;

	for (i=0;i<vols;++i)
	{
		if (volumes[i].state!=volumes[i].alarmed || volumes[i].state==1)
		{
			iao_alarm *a = &(alarms[alarmcount++]);
			strcpy(a->id, "VOLUME_CAPACITY");
			sprintf(a->resource, "Volume %s", volumes[i].name);
			a->severity = 'M';
			
			if (volumes[i].state==1)
			{
				sprintf(a->description, "Volume %s usage has exceeded %d%% (%d%%)", volumes[i].name, volume_threshold, volumes[i].value);
				a->type = 'A';
			}
			else
			{
				sprintf(a->description, "Volume %s usage is back below %d%% (%d%%)", volumes[i].name, volume_threshold, volumes[i].value);
				a->type = 'C';
			}
		}
	}

	for (i=0;i<METRICSCOUNT;++i)
	{
		if (metricstates[i].name!=NULL && (metricstates[i].state!=metricstates[i].alarmed||metricstates[i].state==1))
		{
			iao_alarm *a = &(alarms[alarmcount++]);			
			strcpy(a->id,metricalarmid[i]);
			strcpy(a->resource, metricresourcename[i]);
			a->severity = metricseverity[i];
			if (metricstates[i].state==1)
			{
				sprintf(a->description, "%s usage has exceeded %d%% (%d%%)", metricresourcename[i], metricstates[i].threshold, metricstates[i].value);
				a->type = 'A';
			}
			else
			{
				sprintf(a->description, "%s usage is back below %d%% (%d%%)", metricresourcename[i], metricstates[i].threshold, metricstates[i].value);
				a->type = 'C';
			}
		}
	}

#ifdef WIN32
    i = 0;
    customthresholdsalarms(alarms+alarmcount, MAX_ALARMS-alarmcount, &i);
    alarmcount+=i;
#endif
	alarmcount+=checkpaththresholds(iao_configvalue("Path_Monitor"), alarms+alarmcount);

	/* Update internal state */
	/* Since using iao_alarms_ifchange, we can always save state */
	
	for (i=0;i<METRICSCOUNT;++i)
	{
		if (metricstates[i].name!=NULL)
		{
			metricstates[i].alarmed = metricstates[i].state;
		}
	}
	for (i=0;i<vols;++i)
	{
		volumes[i].alarmed = volumes[i].state;
	}
#ifdef WIN32
        customthresholdreported();
#endif
	return iao_alarms_ifchange(agent, alarms, alarmcount, verb, x);
}

int checkmetric(metricstate *m, int value)
{
	int state = value>m->threshold;
	m->value = value;
	if (state==m->state) 
	{
		m->consecutive = 0;
	}
	else 
	{
		if ((++(m->consecutive)) >= CONSECUTIVEREADINGS)
		{
			m->state = state;
			m->consecutive = 0;
		}
	}
	return m->state!=m->alarmed;
}

	
int *metrics()
{
	static int ms[METRICSCOUNT];
	int i;
	for (i=0;i<METRICSCOUNT;++i)
	{
		ms[i] = -1;
	}
	if (getmetrics(ms)!=0)
	{
		return NULL;
	}
	return ms;
}

int checkmetrics(int *alarms)
{
	int *ms = metrics();
	int i;
	int change = 0;

	for (i=0;i<METRICSCOUNT;++i)
	{
		if (metricstates[i].name!=NULL)
		{
			if (checkmetric(metricstates+i, ms[i]))
			{
				change = 1;
			}
			if (metricstates[i].state!=0)
			{
				(*alarms)++;
			}
		}
	}
	return change;
}

int ignorevolume(char *vname)
{
	char buff[256];
	sprintf(buff, "\n%s\n", vname);
	return strstr(ignorevolumes, buff)!=NULL;
}


int agentmain(iao_agentinfo *agent, char *verb, char *x, int mode)
{
	int i;
	int alarms = 0;

	/* Load Configuration */
	if (mode&MODE_CONFIG_CHANGED)
	{
		for (i=0;i<METRICSCOUNT;++i)
		{
			if (metricstates[i].name!=NULL)
			{
				char buff[256];
				sprintf(buff, "%s_Threshold", metricstates[i].name);
				metricstates[i].threshold = atoi(iao_configvalue(buff));
			}
		}

		if (vols>0)
		{
			volume_threshold = atoi(iao_configvalue("Volume_Threshold"));
			ignorevolumes[0] = '\n';
			strncpy(ignorevolumes+1, iao_configvalue("Ignore_Volumes"), 4092);
			ignorevolumes[4092] = 0;
			strcat(ignorevolumes, "\n");
		}
		else
		{
			ignorevolumes[0]=0;
		}
	}

	if (time(NULL)-lastload>60*60*24*3) /* every 3 days, reload drives */
	{
		initvolumes();
	}

	/* Find out our current state */
	for (i=0;i<vols;++i)
	{
			
		if (ignorevolume(volumes[i].name))
		{
			volumes[i].state = 0;
		}
		else
		{
			int pcnt = volumeusage(volumes[i].name);
			volumes[i].value = pcnt;
			volumes[i].state = pcnt > volume_threshold;
			if (volumes[i].state!=0)
			{
				alarms++;
			}
		}
	}

	checkmetrics(&alarms);

#ifdef WIN32
    checkcustomthresholds(&alarms, iao_configvalue("Custom_Thresholds"), mode&MODE_CONFIG_CHANGED);
#endif

	return reportstate(agent, verb, x);
}


agentmetadata getagentinfo()
{
	static agentmetadata meta;
	static iao_confparam agentparams[METRICSCOUNT+30];
	int *ms  = metrics();
	int metricscount=0;
	int i=0;

	initvolumes();
	if (vols>0)
	{
		agentparams[metricscount].name = "Volume_Threshold";
		agentparams[metricscount].type = "Integer(%)";
		agentparams[metricscount].defvalue = "85";
		metricscount++;
	}
	for (i=0;i<METRICSCOUNT;++i)
	{
		if (ms[i]!=-1)
		{
			char *p = (char *)malloc(256);
			sprintf(p, "%s_Threshold", metricnames[i]);
			agentparams[metricscount].name = p;
			agentparams[metricscount].type = "Integer(%)";
			agentparams[metricscount].defvalue = metricdefaultthreshold[i];
			metricstates[i].name = metricnames[i];
			metricstates[i].alarmed = -1;
			metricstates[i].state = 0;
			metricstates[i].threshold = atoi(metricdefaultthreshold[i]);
			metricstates[i].consecutive = 0;
			metricscount++;
		}
		else
		{
			metricstates[i].name = NULL;
		}
	}

	if (vols>0) 
	{
		agentparams[metricscount].name = "Ignore_Volumes";
		agentparams[metricscount].defvalue = "";
		agentparams[metricscount].type = "Text";
		metricscount++;
	}
#ifdef WIN32
    agentparams[metricscount].name = "Custom_Thresholds";
    agentparams[metricscount].defvalue = "";
    agentparams[metricscount].type = "Text"; 
    metricscount++;
#endif

	agentparams[metricscount].name = "Path_Monitor";
	agentparams[metricscount].defvalue = "";
    	agentparams[metricscount].type = "Text"; 
    	metricscount++;
		

	agentparams[metricscount].defvalue=NULL;
	agentparams[metricscount].name=NULL;
	agentparams[metricscount].type=NULL;
	

	meta.Name = "ServerOn";
	meta.name = "serveron";
	/* When writing your own private agent don't add the ruon prefix to its type name. */
	/* This way you are certain to maintain name uniqueness. */
	meta.type = "ruon.ServerOn";
#ifdef Darwin
        meta.version = "1.9";
#else
        meta.version = "1.8";
#endif
	meta.agentparams = agentparams;
	meta.resourceparams = NULL;
	meta.agentregister = NULL;
	meta.agentinit = agentinit;
	meta.agentmain = agentmain;
	return meta;
}

