#ifdef WIN32
#include <windows.h>
#include <pdh.h>
#include <pdhmsg.h>
#include <time.h>
#include <stdio.h>

#pragma warning(disable:4996)

#include "common.h"
#include "ciao.h"
#include "system.h"

#define MAX_THRESHOLDS 255
#define MAX_RESOURCE_NAME 512

#define OK 0
#define THRESHOLDALARM 1
#define PARSEERROR 2
#define READCOUNTERERROR 3
#define FORMATTEDVALUEERROR 4

static char *internalerrors[] = {"","","Failed parsing custom threshold", "Failed querying custom threshold", "Error formatting custom threshold value"};

typedef struct {
    char resource[MAX_RESOURCE_NAME];
    int greaterthan;
    double threshold;
    double value;
    int reported;
    int actual;
    PDH_HCOUNTER counter;
    char severity;
} customthreshold;


static customthreshold *thresholds = NULL;
static int thresholdscount = 0;
static PDH_HQUERY pdhq=NULL;
static int errorcollecting = 0;


/******* parsing **********/
static int iswhitespace(char c)
{
    static char *whitespace = " \t\r\n";
    return strchr(whitespace, c)!=NULL;
}
static int isempty(char *p)
{
    while (*p!=0)
    {
        if (!iswhitespace(*(p++)))
        {
            return 0;
        }
    }
    return 1;
}

static void trim(char *s)
{
    char * p = s;
    while (*p!=0&&iswhitespace(*p))
    {
        p++;
    }
    if (p!=s)
    {
        strcpy(s, p);
    }
    p = s + strlen(s)-1;
    while (p!=s && iswhitespace(*p))
    {
        p--;
    }
    *(p+1) = 0;
}

static void parse3(char *expression, customthreshold *threshold)
{
    char *p;

    p = strstr(expression, "<");
    threshold->reported = -1;
    threshold->actual = OK;
    if (p==NULL)
    {
        p = strstr(expression, ">");
    }
    if (p==NULL)
    {
        strncpy(threshold->resource, expression, MAX_RESOURCE_NAME);
        threshold->resource[MAX_RESOURCE_NAME-1] = 0;
        threshold->actual = 2;
        return;
    }
    threshold->greaterthan = (*p=='>');
    *p = 0;
    p++;
    trim(p);
    trim(expression);

    threshold->severity = 'M';
    if (strlen(p)>0)
    {
       char sev = p[strlen(p) - 1];  
       if (sev=='C' || sev=='m')
       {
          threshold->severity = sev;
       }
       if (sev=='c')
       {
           threshold->severity = 'C';
       }
    }


    if (expression[0]!='\\')
    {
        threshold->resource[0] = '\\';
        strncpy(threshold->resource+1, expression, MAX_RESOURCE_NAME-1);
    }
    else
    {
        if (expression[1]=='\\') /* no remote yet */
        {
            expression[1] = ' ';
        }
        strncpy(threshold->resource, expression, MAX_RESOURCE_NAME);
    }
    threshold->resource[MAX_RESOURCE_NAME-1] = 0;

    threshold->threshold = atof(p);

}

static void parse2(char **expressions, int count)
{
    int i;
    if (thresholds!=NULL)
    {
        free(thresholds);
    }
    thresholds = (customthreshold *) malloc(sizeof(customthreshold)*count);
    for (i=0;i<count;++i)
    {
        parse3(expressions[i],thresholds+i);
    }
    thresholdscount = count;
}
static void parse1(char *customtext)
{
    char *expressions[MAX_THRESHOLDS];
    char *p = customtext;
    int found = 0;

    while (found<MAX_THRESHOLDS)
    {
        char *t = p;
        char *nl = strstr(t, "\n");
        if (nl!=NULL)
        {
            *nl = 0;
            p+=strlen(t)+1;
        }
        if (!isempty(t)) 
        {
            expressions[found++] = t;
        }
        if (nl==NULL)
        {
            break;
        }
    }
    parse2(expressions, found);
}

static void parse(char *customtext)
{
    char *buff = malloc(strlen(customtext)+1);
    strcpy(buff, customtext);
    parse1(buff);
    free(buff);
}




/********* pdh ****************/
static void createcounters()
{
    int i;
    int hasone = 0;
    iao_logmsg("createcounters: creating counters...");
    for (i=0;i<thresholdscount;++i)
    {
        customthreshold *threshold = thresholds+i;
        if (threshold->actual==OK)
        {
            char buff[1024];
            sprintf(buff, "|%s|%d", threshold->resource, strlen(threshold->resource));
            iao_logmsg(buff);
            if (PdhAddCounter(pdhq, threshold->resource, 0, &threshold->counter) != ERROR_SUCCESS) 
            {

                iao_logsysmsg("Error adding custom threshold");
                threshold->actual = READCOUNTERERROR;
                continue;
            }
            hasone = 1;
        }
        else
        {
            iao_logmsg("createcounters: counter skipped");
        }
    }
    if (hasone)
    {
        PdhCollectQueryData(pdhq);
        Sleep(10000);
    }
}
static int updateactual(int *alarms)
{
    int i;
    int change = 0;
    for (i=0;i<thresholdscount;++i)
    {
        customthreshold *threshold = thresholds+i;
        if (threshold->actual==OK || threshold->actual==THRESHOLDALARM || threshold->actual==FORMATTEDVALUEERROR)
        {
            PDH_FMT_COUNTERVALUE countervalue;
            if (PdhGetFormattedCounterValue(threshold->counter, PDH_FMT_DOUBLE, NULL, &countervalue)!= ERROR_SUCCESS || countervalue.CStatus!=ERROR_SUCCESS) 
            {
                iao_logsysmsg("Error getting counter value");
                threshold->actual = FORMATTEDVALUEERROR;
            }
            else
            {
                threshold->value = countervalue.doubleValue;
                threshold->actual = threshold->greaterthan ?
                    (threshold->value > threshold->threshold?THRESHOLDALARM:OK):
                    (threshold->value < threshold->threshold?THRESHOLDALARM:OK);
            }
            if (threshold->actual!=threshold->reported)
            {
                change = 1;
            }
        }
        if (threshold->actual!=OK)
        {
            (*alarms)++;
        }
    }
    return change;
}
static void descriptionformat(char *buff, customthreshold *threshold)
{
    int tt = (int) threshold->threshold;
    int tv = (int) threshold->value;

    if ((double)tt==threshold->threshold && (double)tv==threshold->value)
    {
        char *format = threshold->greaterthan?"Resource exceeded %d (%d)": "Resource is below %d (%d)";
        sprintf(buff, format, tt, tv);
    }
    else
    {
        char *format = threshold->greaterthan?"Resource exceeded %.2f (%.2f)": "Resource is below %.2f (%.2f)";
        sprintf(buff, format, threshold->threshold, threshold->value );
    }
}

/********** public **********/
int checkcustomthresholds(int *alarms, char *customtext, int updateconfig)
{
    PDH_STATUS ret;

    if (customtext==NULL)
    {
        *alarms = 0;
        return 0;
    }

    if (updateconfig )
    {
        if (pdhq!=NULL)
        {
            PdhCloseQuery(pdhq);
        }
        thresholdscount = 0;
        parse(customtext);
        if (thresholdscount>0)
        {
            if (PdhOpenQuery(NULL, 0, &pdhq) != ERROR_SUCCESS)
            {
                pdhq = NULL;
                iao_logsysmsg("Error while openning customthresholds query");
                *alarms = 1;
                return 1;
            }
            else
            {
                createcounters();
            }
        }
    }

    if (pdhq==NULL)
    {
        *alarms = 1;
        return 0;
    }

    ret = PdhCollectQueryData(pdhq);
    if (ret!=ERROR_SUCCESS)
    {
        if (ret==PDH_INVALID_HANDLE) 
        {
            iao_logsysmsg("Error collecting custom threholds data - PDH_INVALID_HANDLE");
        }
        else if (ret==PDH_NO_DATA)
        {
            iao_logsysmsg("Error collecting custom threholds data - PDH_NO_DATA");
        }
        else
        {
            char buff[256];
            sprintf(buff, "Error collecting custom threholds data - %d", ret);
            iao_logsysmsg(buff);
        }
        *alarms=1;
        return 0==errorcollecting++;
    }
    errorcollecting = 0;

    return updateactual(alarms);
}

void customthresholdsalarms(iao_alarm *alarms, int max, int *count)
{
    int i = 0;

    if (pdhq==NULL)
    {
        if (thresholdscount>0)
        {
            sprintf(alarms[0].description, "Couldn't open custom counters system");
            sprintf(alarms[0].id, "CUSTOM_THRESHOLD_INTERNAL");
            sprintf(alarms[0].resource, "PDH_SYSTEM");
            alarms[0].severity = 'm';
            alarms[0].type = 'A';
            (*count)++;
        }
        return;
    }

    if (errorcollecting)
    {
        sprintf(alarms[0].description, "Error collecting custom counters");
        sprintf(alarms[0].id, "CUSTOM_THRESHOLD_INTERNAL");
        sprintf(alarms[0].resource, "PDH_SYSTEM");
        alarms[0].severity = 'm';
        alarms[0].type = 'A';
        (*count)++;
        return;
    }


    while(i<thresholdscount && *count<max)
    {
        customthreshold *threshold = thresholds+(i++);
        switch(threshold->actual)
        {
        case OK:
            continue;
        case THRESHOLDALARM:
            descriptionformat(alarms[*count].description, threshold);
            sprintf(alarms[*count].id, "CUSTOM_THRESHOLD_CROSS");
            sprintf(alarms[*count].resource, threshold->resource);
            alarms[*count].severity = threshold->severity;
            alarms[*count].type = 'A';
            break;
        case PARSEERROR:
        case READCOUNTERERROR:
        case FORMATTEDVALUEERROR:
            sprintf(alarms[*count].description, internalerrors[threshold->actual]);
            sprintf(alarms[*count].id, "CUSTOM_THRESHOLD_INTERNAL");
            sprintf(alarms[*count].resource, threshold->resource);
            alarms[*count].severity = 'm';
            alarms[*count].type = 'A';
            break;
        }
        (*count)++;
    }
}
void customthresholdreported()
{
    int i;
    for (i=0;i<thresholdscount;++i)
    {
        customthreshold *threshold = thresholds+i;
        threshold->reported = threshold->actual;
    }
}
#endif
